@@ -11,7 +11,7 @@ use feature 'current_sub';
 use experimental 'signatures', 'lexical_subs', 'postderef';
 
 #<<< no tidy
-our $VERSION = '0.09'; # TRIAL
+our $VERSION = '0.10'; # TRIAL
 #>>>
 
 use base 'Exporter::Tiny';
@@ -36,20 +36,29 @@ use constant {
     VISIT_ALL       => 0b111,
 };
 use constant {
-    RESULT_RETURN            => 0,
-    RESULT_CONTINUE          => 1,
-    RESULT_REVISIT_CONTAINER => 2,
-    RESULT_REVISIT_ELEMENT   => 3,
-    RESULT_STOP_DESCENT      => 4,
+    RESULT_NULL              => 0b00000,
+    RESULT_RETURN            => 0b00001,
+    RESULT_CONTINUE          => 0b00010,
+    RESULT_REVISIT_CONTAINER => 0b00100,
+    RESULT_REVISIT_ELEMENT   => 0b01000,
+    RESULT_STOP_DESCENT      => 0b10000,
 };
 
-use constant { PASS_VISIT_ELEMENT => 1, PASS_REVISIT_ELEMENT => 2 };
+use constant {
+    PASS_VISIT_ELEMENT   => 0,
+    PASS_REVISIT_ELEMENT => 1,
+};
 
 our %EXPORT_TAGS = (
     funcs   => [qw( visit )],
-    results => [
-        qw( RESULT_RETURN RESULT_CONTINUE RESULT_REVISIT_CONTAINER
-          RESULT_REVISIT_ELEMENT RESULT_STOP_DESCENT ),
+    results => [ qw(
+          RESULT_NULL
+          RESULT_RETURN
+          RESULT_CONTINUE
+          RESULT_REVISIT_CONTAINER
+          RESULT_REVISIT_ELEMENT
+          RESULT_STOP_DESCENT
+        ),
     ],
     cycles    => [qw( CYCLE_DIE CYCLE_CONTINUE CYCLE_TRUNCATE )],
     visits    => [qw( VISIT_ARRAY VISIT_HASH VISIT_CONTAINER VISIT_LEAF VISIT_ALL )],
@@ -98,6 +107,7 @@ my sub _visit ( $node, $code, $context, $cycle, $visit, $meta ) {
     defer { delete $meta->{seen}{$refaddr} }
 
     my %meta = $meta->%*;
+    @meta{ 'visit', 'idx' } = ( 0, -1 );
     $meta{container} = $node;
 
     # deal with bare next in $code body
@@ -116,7 +126,12 @@ my sub _visit ( $node, $code, $context, $cycle, $visit, $meta ) {
   SCAN: {
         last unless --$revisit_limit;
 
-        my $idx_arr = do {
+        $meta{visit}++;
+        $meta{idx} = -1;
+
+        my $revisit_container = !!0;
+
+        my $kydx_arr = do {
 
             if ( $is_hashref ) {
                     $sort_key_mode == 0 ? $key_sort->( [ keys $node->%* ] )
@@ -132,12 +147,13 @@ my sub _visit ( $node, $code, $context, $cycle, $visit, $meta ) {
             }
         };
 
-        for my $idx ( $idx_arr->@* ) {
+        for my $kydx ( $kydx_arr->@* ) {
+            $meta{idx}++;
 
-            push $path->@*, $idx;
+            push $path->@*, $kydx;
             defer { pop $path->@* }
 
-            my $vref = \( $is_hashref ? $node->{$idx} : $node->[$idx] );
+            my $vref = \( $is_hashref ? $node->{$kydx} : $node->[$kydx] );
 
             my $visit_element
               = is_plain_hashref( $$vref )  ? $visit_hash
@@ -148,18 +164,22 @@ my sub _visit ( $node, $code, $context, $cycle, $visit, $meta ) {
 
             $meta{pass} = PASS_VISIT_ELEMENT;
             if ( $visit_element
-                and ( my $result = $code->( $idx, $vref, $context, \%meta ) ) != RESULT_CONTINUE )
+                and ( my $result = $code->( $kydx, $vref, $context, \%meta ) ) != RESULT_CONTINUE )
             {
-                redo SCAN  if $result == RESULT_REVISIT_CONTAINER;
-                return !!0 if $result == RESULT_RETURN;
-                next       if $result == RESULT_STOP_DESCENT;        # this works for both leaves and containers
+                # immediate rescan if explicitly set to value,
+                # otherwise it will happen after the container is
+                # completely visited
+                redo SCAN if $result == RESULT_REVISIT_CONTAINER;
 
-                if ( $result == RESULT_REVISIT_ELEMENT ) {
-                    $revisit_element = !!1;
-                }
-                elsif ( $result != RESULT_CONTINUE ) {
-                    croak( "unknown return value from visit: $result" );
-                }
+                return !!0 if $result == RESULT_RETURN;
+                $revisit_container = $result & RESULT_REVISIT_CONTAINER;
+
+                next if $result & RESULT_STOP_DESCENT;    # this works for both leaves and containers
+
+                $revisit_element = $result & RESULT_REVISIT_ELEMENT;
+
+                croak( "unknown return value from visit: $result" )
+                  if !$revisit_element && !$result & RESULT_CONTINUE;
             }
 
             next unless is_plain_refref( $vref );
@@ -167,25 +187,25 @@ my sub _visit ( $node, $code, $context, $cycle, $visit, $meta ) {
             my $ref = $vref->$*;
             if ( is_plain_arrayref( $ref ) || is_plain_hashref( $ref ) ) {
                 __SUB__->( $ref, $code, $context, $cycle, $visit, \%meta ) || return !!0;
-
                 if ( $revisit_element ) {
                     $meta{pass} = PASS_REVISIT_ELEMENT;
-                    my $result = $code->( $idx, $vref, $context, \%meta );
+                    my $result = $code->( $kydx, $vref, $context, \%meta );
                     return !!0 if $result == RESULT_RETURN;
-                    next       if $result == RESULT_CONTINUE;
-                    croak( "unknown return value from visit: $result" );
+                    croak( "unknown return value from visit: $result" )
+                      if $result & ~( RESULT_CONTINUE | RESULT_REVISIT_CONTAINER );
+                    $revisit_container |= $result & RESULT_REVISIT_CONTAINER;
                 }
             }
         }
+        redo SCAN if $revisit_container;
     }
-    croak( "exceeded limit ($meta->{revisit_limit}) on revisiting containers" )
+    croak( "exceeded limit ($meta{revisit_limit}) on revisiting containers" )
       unless $revisit_limit;
 
     return !!1;
 }
+
 ## critic (Subroutines::ProhibitManyArgs  Subroutines::ProhibitExcessComplexity)
-
-
 sub visit ( $struct, $callback, %opts ) {
 
     is_coderef( $callback )
@@ -277,7 +297,7 @@ __END__
 
 =pod
 
-=for :stopwords Diab Jerius Smithsonian Astrophysical Observatory selectable
+=for :stopwords Diab Jerius Smithsonian Astrophysical Observatory selectable idx
 
 =head1 NAME
 
@@ -285,26 +305,30 @@ CXC::Data::Visitor - Invoke a callback on every element at every level of a data
 
 =head1 VERSION
 
-version 0.09
+version 0.10
 
 =head1 SYNOPSIS
 
-  use CXC::Data::Visitor 'visit', 'RESULT_CONTINUE';
+ use CXC::Data::Visitor 'visit', 'RESULT_CONTINUE';
+ 
+ my %struct = (
+     fruit => {
+         berry  => 'purple',
+         apples => [ 'fuji', 'macoun' ],
+     } );
+ 
+ visit(
+     \%struct,
+     sub ( $kydx, $vref, @ ) {
+         $vref->$* = 'blue' if $kydx eq 'berry';
+         return RESULT_CONTINUE;
+     } );
+ 
+ say $struct{fruit}{berry}
 
-  my %struct = (
-      fruit => {
-          berry  => 'purple',
-          apples => [ 'fuji', 'macoun' ],
-      } );
+results in
 
-  visit(
-      \%struct,
-      sub ( $kydx, $vref, @ ) {
-          $vref->$* = 'blue' if $kydx eq 'berry';
-          return RESULT_CONTINUE;
-      } );
-
-  say $struct{fruit}{berry}    # 'blue'
+ blue
 
 =head1 DESCRIPTION
 
@@ -381,7 +405,7 @@ collected metadata. See L</Metadata>.
 
 =head2 Options
 
-C<visit> may b passed the following options:
+C<visit> may be passed the following options:
 
 =over
 
@@ -430,17 +454,18 @@ than, equal to, or greater than that of the second argument.
 
 =item revisit_limit
 
-If L</$callback> returns B<RESULT_REVISIT_CONTAINER> the element's parent container
-is re-scanned for its elements and revisited.  To avoid an inadvertent
-infinite loop, an exception is thrown if the parent container is revisited more
-than this number of times.  It defaults to 10;  Set it to C<0> to indicate no limit.
+If L</$callback> returns B<RESULT_REVISIT_CONTAINER> the element's
+parent container is re-scanned for its elements and revisited.  To
+avoid an inadvertent infinite loop, an exception is thrown if the
+parent container is revisited more than this number of times.  It
+defaults to 10; Set it to C<0> to indicate no limit.
 
 =back
 
 =head2 Callback
 
-C<visit> invokes C<$callback> on selected elements of
-C<$struct> (see L</Element Filters>). C<$callback> is invoked as
+C<visit> invokes C<$callback> on selected elements of C<$struct> (see
+L</Element Filters>). C<$callback> is invoked as
 
   $directive = $callback->( $kydx, $vref, $context, \%metadata );
 
@@ -477,22 +502,22 @@ See L</Metadata>
 
 =head2 Traversal Order
 
-C<$struct> is traversed in depth-first order.
-At each level elements are traversed in sorted order.  For hashes,
-this is alphabetical, for arrays, numerical. The defaults may be
-changed with the L</key_sort> and L</idx_sort> options.
+C<$struct> is traversed in depth-first order.  At each level elements
+are traversed in sorted order.  For hashes, this is alphabetical, for
+arrays, numerical. The defaults may be changed with the L</key_sort>
+and L</idx_sort> options.
 
 The default traversal order for the structure in the L</SYNOPSIS> is
 
-  +---------------------------+-------------------------+
-  | Path                      | Value                   |
-  +---------------------------+-------------------------+
-  | $struct{fruit}            | \$struct{fruit}         |
-  | $struct{fruit}{apples}    | \$struct{fruit}{apples} |
-  | $struct{fruit}{apples}[0] | fuji                    |
-  | $struct{fruit}{apples}[1] | macoun                  |
-  | $struct{fruit}{berry}     | purple                  |
-  +---------------------------+-------------------------+
+ +---------------------------+-------------------------+-----+
+ | path                      | value                   | idx |
+ +---------------------------+-------------------------+-----+
+ | $struct{fruit}            | \$struct{fruit}         | 0   |
+ | $struct{fruit}{apples}    | \$struct{fruit}{apples} | 0   |
+ | $struct{fruit}{apples}[0] | fuji                    | 0   |
+ | $struct{fruit}{apples}[1] | macoun                  | 1   |
+ | $struct{fruit}{berry}     | purple                  | 1   |
+ +---------------------------+-------------------------+-----+
 
 Containers that can be reached multiple times without cycling, e.g.
 
@@ -515,6 +540,16 @@ indicating what C<visit> should do next.
 
 continue traversing as normal.
 
+ +---------------------------+-------------------------+-----+
+ | path                      | value                   | idx |
+ +---------------------------+-------------------------+-----+
+ | $struct{fruit}            | \$struct{fruit}         | 0   |
+ | $struct{fruit}{apples}    | \$struct{fruit}{apples} | 0   |
+ | $struct{fruit}{apples}[0] | fuji                    | 0   |
+ | $struct{fruit}{apples}[1] | macoun                  | 1   |
+ | $struct{fruit}{berry}     | purple                  | 1   |
+ +---------------------------+-------------------------+-----+
+
 =item RESULT_RETURN
 
 return immediately to the caller of C<visit>.
@@ -529,34 +564,33 @@ For example, If C<RESULT_STOP_DESCENT> is returned when
 C<$struct{fruit}{apples}> is traversed, the traversal would look like
 this:
 
-  +------------------------+--------=----------------+
-  | Path                   | Value                   |
-  +------------------------+----------=--------------+
-  | $struct{fruit}         | \$struct{fruit}         |
-  | $struct{fruit}{apples} | \$struct{fruit}{apples} |
-  | $struct{fruit}{berry}  | purple                  |
-  +------------------------+-------------------------+
+ +------------------------+-------------------------+-----+
+ | path                   | value                   | idx |
+ +------------------------+-------------------------+-----+
+ | $struct{fruit}         | \$struct{fruit}         | 0   |
+ | $struct{fruit}{apples} | \$struct{fruit}{apples} | 0   |
+ | $struct{fruit}{berry}  | purple                  | 1   |
+ +------------------------+-------------------------+-----+
 
 =item RESULT_REVISIT_CONTAINER
 
-Stop processing the current container, and start traversing
+Immediately stop processing the current container, and start traversing
 it again.
 
 For example, if C<RESULT_REVISIT_CONTAINER> is returned the
-first time C<$struct{fruit}{apples}[1]> is traversed, the
+first time C<$struct{fruit}{apples}[0]> is traversed, the
 traversal would look like this:
 
-  +---------------------------+-------------------------+
-  | Path                      | Value                   |
-  +---------------------------+-------------------------+
-  | $struct{fruit}            | \$struct{fruit}         |
-  | $struct{fruit}{apples}    | \$struct{fruit}{apples} |
-  | $struct{fruit}{apples}[0] | fuji                    |
-  | $struct{fruit}{apples}[1] | macoun                  |
-  | $struct{fruit}{apples}[0] | fuji                    |
-  | $struct{fruit}{apples}[1] | macoun                  |
-  | $struct{fruit}{berry}     | purple                  |
-  +---------------------------+-------------------------+
+ +---------------------------+-------------------------+-----+-------+
+ | path                      | value                   | idx | visit |
+ +---------------------------+-------------------------+-----+-------+
+ | $struct{fruit}            | \$struct{fruit}         | 0   | 1     |
+ | $struct{fruit}{apples}    | \$struct{fruit}{apples} | 0   | 1     |
+ | $struct{fruit}{apples}[0] | fuji                    | 0   | 1     |
+ | $struct{fruit}{apples}[0] | fuji                    | 0   | 2     |
+ | $struct{fruit}{apples}[1] | macoun                  | 1   | 2     |
+ | $struct{fruit}{berry}     | purple                  | 1   | 1     |
+ +---------------------------+-------------------------+-----+-------+
 
 To avoid inadvertent infinite loops, the number of revisits
 during a traversal of a container is limited (see L</revisit_limit>).
@@ -565,9 +599,11 @@ limit is reset for each traversal.
 
 =item RESULT_REVISIT_ELEMENT
 
-If the element is a container, it will be revisited
+If the element is a container, L</$callback> will be invoked with it
 after its contents are traversed.
 
+If the element is not a container, the result is the same as L</RESULT_CONTINUE>.
+
 During the initial visit
 
   $metadata->{pass} & PASS_VISIT_ELEMENT
@@ -576,9 +612,79 @@ will be true.  During the followup visit
 
   $metadata->{pass} & PASS_REVISIT_ELEMENT
 
-will be true. During this visit, L</$callback> must return either
-B<RESULT_RETURN> or B<RESULT_CONTINUE>, otherwise an exception
-will be thrown.
+will be true. During this visit, L</$callback> must return one of the following:
+
+  RESULT_RETURN
+  RESULT_CONTINUE
+  RESULT_CONTINUE | RESULT_REVISIT_CONTAINER
+
+otherwise an exception will be thrown.
+
+For example, If C<RESULT_REVISIT_ELEMENT> is returned when
+C<$struct{fruit}{apples}> is traversed, the traversal would look like
+this:
+
+ +---------------------------+-------------------------+-----+----------------------+
+ | path                      | value                   | idx | pass                 |
+ +---------------------------+-------------------------+-----+----------------------+
+ | $struct{fruit}            | \$struct{fruit}         | 0   | PASS_VISIT_ELEMENT   |
+ | $struct{fruit}{apples}    | \$struct{fruit}{apples} | 0   | PASS_VISIT_ELEMENT   |
+ | $struct{fruit}{apples}[0] | fuji                    | 0   | PASS_VISIT_ELEMENT   |
+ | $struct{fruit}{apples}[1] | macoun                  | 1   | PASS_VISIT_ELEMENT   |
+ | $struct{fruit}{apples}    | \$struct{fruit}{apples} | 0   | PASS_REVISIT_ELEMENT |
+ | $struct{fruit}{berry}     | purple                  | 1   | PASS_VISIT_ELEMENT   |
+ +---------------------------+-------------------------+-----+----------------------+
+
+=back
+
+Some directives can be mixed with L</RESULT_REVISIT_CONTAINER> and
+L</RESULT_REVISIT_ELEMENT> by supplying a binary with them.
+
+=over
+
+=item RESULT_STOP_DESCENT | RESULT_REVISIT_CONTAINER
+
+Modify the next processing step as performed by L</RESULT_STOP_DESCENT>
+
+Additionally, revisit the container after the rest of the container
+has been visited.
+
+For example, if C<RESULT_STOP_DESCENT | RESULT_REVISIT_CONTAINER> is
+returned when C<$struct{fruit}{apples}> is traversed when
+C<$metadata->{visit} ==1>, the traversal would look like
+
+ +------------------------+-------------------------+-----+-------+
+ | path                   | value                   | idx | visit |
+ +------------------------+-------------------------+-----+-------+
+ | $struct{fruit}         | \$struct{fruit}         | 0   | 1     |
+ | $struct{fruit}{apples} | \$struct{fruit}{apples} | 0   | 1     |
+ | $struct{fruit}{berry}  | purple                  | 1   | 1     |
+ | $struct{fruit}{apples} | \$struct{fruit}{apples} | 0   | 2     |
+ | $struct{fruit}{berry}  | purple                  | 1   | 2     |
+ +------------------------+-------------------------+-----+-------+
+
+=item RESULT_CONTINUE | RESULT_REVISIT_CONTAINER
+
+Modify the next processing step as performed by L</RESULT_CONTINUE>
+
+Additionally, revisit the container after the rest of the container
+has been visited.
+
+For example, if C<RESULT_CONTINUE | RESULT_REVISIT_CONTAINER> is
+returned when C<$callback> is first passed
+C<$struct{fruit}{apples}[0]>, the traversal would look like
+
+ +---------------------------+-------------------------+-----+-------+
+ | path                      | value                   | idx | visit |
+ +---------------------------+-------------------------+-----+-------+
+ | $struct{fruit}            | \$struct{fruit}         | 0   | 1     |
+ | $struct{fruit}{apples}    | \$struct{fruit}{apples} | 0   | 1     |
+ | $struct{fruit}{apples}[0] | fuji                    | 0   | 1     |
+ | $struct{fruit}{apples}[1] | macoun                  | 1   | 1     |
+ | $struct{fruit}{apples}[0] | fuji                    | 0   | 2     |
+ | $struct{fruit}{apples}[1] | macoun                  | 1   | 2     |
+ | $struct{fruit}{berry}     | purple                  | 1   | 1     |
+ +---------------------------+-------------------------+-----+-------+
 
 =back
 
@@ -608,8 +714,20 @@ current element.
 
 =item B<pass>
 
-A constant indicating the current visit pass through an element.
-See L</RESULT_REVISIT_CONTAINER>.
+A constant indicating the current visit pass of an element.
+See L</RESULT_REVISIT_ELEMENT>.
+
+=item B<visit>
+
+A unary-based counter indicating the number of times a container has
+been scanned in a single visit.  This will be greater than C<1> if the
+L</RESULT_REVISIT_CONTAINER> directive has been applied.
+
+=item B<idx>
+
+A zero-based index indicating the order of the element in its container.
+Ordering depends upon how container elements are sorted; see
+L</key_sort> and L</idx_sort>.
 
 =back
 
